// ZeroTier identity function is not memory hard by Michal Zielinski <michal@zielinscy.org.pl>
#include "Identity.hpp"
#include <cassert>
#include "Constants.hpp"
#include "Identity.hpp"
#include "SHA512.hpp"
#include "Salsa20.hpp"
#include "Utils.hpp"

using namespace ZeroTier;

#define ZT_IDENTITY_GEN_HASHCASH_FIRST_BYTE_LESS_THAN 17
#define ZT_IDENTITY_GEN_MEMORY 2097152

void print_hex(unsigned char* ch, int l) {
    for (int i=0; i < l; i ++) {
        if (i % 8 == 0 && i != 0) printf("-");
        printf("%02x", ch[i]);
    }
    printf("\n");
}

// A memory-hard composition of SHA-512 and Salsa20 for hashcash hashing
bool computeMemoryHardHash(const void *publicKey, unsigned int publicKeyBytes, void *digest, void *genmem, int key_pos)
{
    // Digest publicKey[] to obtain initial digest
    SHA512::hash(digest,publicKey,publicKeyBytes);

    const int count = (ZT_IDENTITY_GEN_MEMORY / sizeof(uint64_t));

    // Initialize genmem[] using Salsa20 in a CBC-like configuration since
    // ordinary Salsa20 is randomly seekable. This is good for a cipher
    // but is not what we want for sequential memory-harndess.

    memset(genmem,0,ZT_IDENTITY_GEN_MEMORY);
    Salsa20 s20(digest,256,(char *)digest + 32);
    s20.encrypt20((char *)genmem,(char *)genmem,64);

    for(unsigned long i=64;i<ZT_IDENTITY_GEN_MEMORY;i+=64) {
        memcpy((char*)genmem + i, (char*)genmem + (i - 64), 64);
        s20.encrypt20((char *)genmem + i, (char *)genmem + i, 64);
    }
    //printf("hard: "); //print_hex((uint8_t*)genmem + ZT_IDENTITY_GEN_MEMORY - 64, 64);
    //print_hex((uint8_t*)genmem + ZT_IDENTITY_GEN_MEMORY - 64 - 64, 64);

    //printf("key is initially: ");
    //print_hex((uint8_t*)genmem + key_pos * 8, 8);

    bool good = true;

    // Render final digest using genmem as a lookup table
    for(unsigned long i=0;i<count;) {
        unsigned long idx1 = (unsigned long)(Utils::ntoh(((uint64_t *)genmem)[i++]) % (64 / sizeof(uint64_t)));
        unsigned long idx2 = (unsigned long)(Utils::ntoh(((uint64_t *)genmem)[i++]) % (ZT_IDENTITY_GEN_MEMORY / sizeof(uint64_t)));
        uint64_t tmp = ((uint64_t *)genmem)[idx2];

        //#define LOG
        if (i != count) {
            if (idx2 >= count - 2 || idx2 == key_pos)
                good = false;
        }

#ifdef LOG
        if (i == count) {
            printf("final, idx2: %d, expected %d\n", (int)idx2, key_pos);
            printf("value: "); print_hex((uint8_t*)genmem + key_pos * 8, 8);
        }
        if (i == count) {
            printf("hop: %d\n", (int)(idx1 % 8));
        }
#endif
        ((uint64_t *)genmem)[idx2] = ((uint64_t *)digest)[idx1];
        ((uint64_t *)digest)[idx1] = tmp;

#ifdef LOG
        if (i == count) {
            printf("hard pre-enc: ");
            print_hex((uint8_t*)digest + 64 - 8, 8);
        }
#endif

        s20.encrypt20(digest,digest,64);

#ifdef LOG
        if (i == count) {
            printf("hard post-enc: ");
            print_hex((uint8_t*)digest + 64 - 8, 8);
        }
#endif
    }

    //printf("hard: ");print_hex((uint8_t*)genmem + ZT_IDENTITY_GEN_MEMORY - 64, 64);
    return good;
}

bool check(const uint8_t* identity) { // length=63
    return identity[0] < 30;
}

bool computeMemoryEasyHash(const void *publicKey, unsigned int publicKeyBytes, void *digest, int& out_key_pos)
{
    const int count = (ZT_IDENTITY_GEN_MEMORY / sizeof(uint64_t));
    const int bcount = (ZT_IDENTITY_GEN_MEMORY / 64);

    SHA512::hash(digest, publicKey, publicKeyBytes);
    Salsa20 s20(digest,256,(char *)digest + 32);
    char mem[64];
    memset(mem, 0, sizeof(mem));
    for(int i=0; i < bcount; i ++) {
        s20.encrypt20((char *)mem, (char *)mem, 64);
    }


    uint64_t key_digit = (uint64_t)Utils::ntoh(*(uint64_t*)((char*)mem + 64 - 8));
    uint64_t idx_digit = (uint64_t)Utils::ntoh(*(uint64_t*)((char*)mem + 64 - 16));
    if (idx_digit % 8 != 7) return false;

    int key_pos = (key_digit % count) * 8;
    out_key_pos = key_pos / 8;

    SHA512::hash(digest, publicKey, publicKeyBytes);
    s20 = Salsa20(digest,256,(char *)digest + 32);
    memset(mem, 0, sizeof(mem));
    for(int i=0; i < bcount - 1; i ++) {
        s20.encrypt20((char *)mem, (char *)mem, 64);
        // mem is now genmem[i*64 : (i+1)*64]
        if (key_pos / 64 == i)
            break;
    }

    uint8_t* key = ((uint8_t*)mem) + (key_pos % 64);
    //printf("pre-enc key:"); print_hex(key, 8);
    uint8_t fullDigest[64];
    memset(fullDigest, 0, 64);
    memcpy(fullDigest + 64 - 8, key, 8);

    // TODO: seek
    char dumb[64];
    SHA512::hash(digest, publicKey, publicKeyBytes);
    s20 = Salsa20(digest,256,(char *)digest + 32);
    for (int i=0; i < bcount + count/2 - 1; i ++) s20.encrypt20(dumb, dumb, 64);

    s20.encrypt20(fullDigest, fullDigest, 64);

    key = fullDigest + 64 - 8;
    if (!check(key + 3)) return false;
    printf("checking candidate: ");
    print_hex(key + 3, 5);

    return true;
}


void generate(int var)
{
    unsigned char digest[64];
    char *genmem = new char[ZT_IDENTITY_GEN_MEMORY];

    C25519::Pair kp;
    char msg[100];
    sprintf(msg, "000000000000000000000000000000000000000000000000000000000000%04x", var);
    assert(strlen(msg) == 64);
    memcpy(kp.priv.data, msg, 64); // 43c74618be
    C25519::_calcPubDH(kp);
    C25519::_calcPubED(kp);

    unsigned char digest2[64];
    int key_pos;
    bool ok = computeMemoryEasyHash(kp.pub.data,(unsigned int)kp.pub.size(),digest2, key_pos);
    if (!ok) return;

    bool hardOk = computeMemoryHardHash(kp.pub.data, (unsigned int)kp.pub.size(), digest, genmem, key_pos);

    if (!hardOk) {
        return;
    }
    printf("found pre-solution\n");

    const unsigned char* addr = digest + 59;

    if (!check(addr)) {
        printf("INTERNAL ERROR: BAD ADDRESS");
        return;
    }

    if (digest[0] >= 17) {
        return;
    }

    printf("SOLUTION: privkey: %04x addr: %02x%02x%02x%02x%02x\n", var, addr[0], addr[1], addr[2], addr[3], addr[4]);
    std::string saddr = Utils::hex(addr, 5) + ":0:" + Utils::hex(kp.pub.data, 64) + ":" + Utils::hex(kp.priv.data, 64);
    Identity ident (saddr);
    if(!ident.locallyValidate()) printf("validation failure!\n");
    printf("id: %s\n", saddr.c_str());

    delete [] genmem;
    printf("------------------------\n");
}

int main() {
    //generate(0);
    for (int i=0; i < 50000; i ++) {
        if (i % 100 == 0) printf("checking %d\n", i);
        generate(i);
    }
}
